@@ -10,9 +10,6 @@ from subprocess import Popen, PIPE
 from os.path import abspath, dirname
 
 
-VERSION = (1, 0, 5, 'final', 0)
-
-
 def git_sha():
     loc = abspath(dirname(__file__))
     p = Popen(
@@ -24,6 +21,9 @@ def git_sha():
     return p.communicate()[0]
 
 
+VERSION = (1, 1, 0, 'final', 0)
+
+
 def get_version(form='short'):
     """
     Return a version string for this package, based on `VERSION`.
@@ -93,6 +93,6 @@ def get_version(form='short'):
     try:
         return versions[form]
     except KeyError:
-        raise TypeError, '"%s" is not a valid form specifier.' % form
+        raise TypeError('"%s" is not a valid form specifier.' % form)
 
 __version__ = get_version('short')
